from sys import stdout
from decimal import Decimal
import numpy as np
from scipy.signal import ZerosPolesGain

# Mark Tillotson, 2017-07-19
# IIR Digital filter compiling to PASM

from utils import factorize_poly, quantize, isclose, max_gain


coeff_bits = 18

def coeff_hex (value):
    if value < 0:
        res = '-'
        value = -value
    else:
        res = '+'
    intpart = value >> coeff_bits
    fract = value - (intpart << coeff_bits)
    res += '%x' % intpart
    res += '.'
    fract += 1<<coeff_bits
    fract <<= (4 - (coeff_bits & 3)) & 3
    res += ('%x' % fract)[1:]
    return res
        

PROP_XTAL = 5.000e6
PROP_CLK = PROP_XTAL * 16
CYCLE_TIME_US = 1.0e6 / PROP_CLK

I2S_XTAL = 6.144e6
I2S_CLK = I2S_XTAL * 16
I2S_CYCLE_TIME_US = 1.0e6 / I2S_CLK


class Outputter:
    def __init__ (self, file):
        self.file = file
        self.cycle_count = 0

    def print (self, string):
        print (string, file = self.file)

    def instr (self, string, label = None, comment = None):
        outstr = "                "
        if label is not None:
            outstr = label + outstr [len(label):]
        outstr += string
        if comment is not None:
            if len(outstr) < 40:
                outstr += ' '*(40-len(outstr))
            outstr += " ' "+comment
        print (outstr, file = self.file)
        self.cycle_count += 4

    def comment (self, comment):
        outstr = ' ' * 40
        outstr += " ' "+comment
        print (outstr, file = self.file)

    def label (self, label, comment = None):
        outstr = "                "
        outstr = label + outstr [len(label):]
        if comment is not None:
            if len(outstr) < 40:
                outstr += ' '*(40-len(outstr))
            outstr += " ' "+comment
        print (outstr, file = self.file)
        

    def big_comment (self, comment):
        print ("{{ " + comment + " }}", file = self.file)

    def blank (self):
        print ("", file = self.file)

    def cycles (self):
        return self.cycle_count


def cycles_desc (cycles):
    return "%i cycles (I2S: %s us, standard: %s us) %i instructs" % (cycles,
                                                                     Decimal (cycles * I2S_CYCLE_TIME_US).quantize (Decimal ('0.001')),
                                                                     Decimal (cycles * CYCLE_TIME_US).quantize (Decimal ('0.001')),
                                                                     cycles//4)

def gen_coeffs (s, ab, n, coeff, out):
    start_cycles = out.cycles()
    neg = coeff < 0.0
    if neg:
        coeff = -coeff
    icoeff = int (round (coeff * (1 << coeff_bits)))
    if icoeff == 2 << coeff_bits:
        icoeff -= 1
    out.label ("coeff%i_%s%i" % (s, ab, n),
               comment = "value += t * %f [ %s hex ]" % (-coeff if neg else coeff, coeff_hex (icoeff)))
    b = 1 << coeff_bits
    if (icoeff & b) != 0:
        out.instr ("%s     value, t" % ("sub" if neg else "add"))
        icoeff -= b
        
    c = 0
    while b != 0:
        while (icoeff & b) == 0:
            c += 1
            b >>= 1
            if b == 0:
                break
        if b == 0:
            break
        d = 0
        while (icoeff & b) != 0:
            d += 1
            b >>= 1

        if d == 1:
            out.instr ("sar     t, #%i" % c)
            out.instr ("%s     value, t" % ("sub" if neg else "add"))
        else:
            if c > 1:
                out.instr ("sar     t, #%i" % (c-1))
            out.instr ("%s     value, t" % ("sub" if neg else "add"))
            out.instr ("sar     t, #%i" % d)
            out.instr ("%s     value, t" % ("add" if neg else "sub"))

        c = 1

    out.instr ("ret", label = "coeff%i_%s%i_ret" % (s, ab, n),
               comment =  cycles_desc (out.cycles() + 4 - start_cycles))
    out.blank ()



def gen_complex_coeffs (s, ab, coeff, out):
    icoeff_re = int (round (coeff.real * (1 << coeff_bits)))
    icoeff_im = int (round (coeff.imag * (1 << coeff_bits)))
    out.label ("coeff%i_%s" % (s, ab),
               comment = "v += t * %s [ %s%sj hex ]" % (coeff, coeff_hex (icoeff_re), coeff_hex (icoeff_im)))
        
    out.instr ("mov     tx, x")
    out.instr ("mov     ty, y")
    gen_complex_coeffs_inner (s, coeff.real, out)

    out.instr ("neg     tx, y")
    out.instr ("mov     ty, x")
    gen_complex_coeffs_inner (s, coeff.imag, out)

    out.instr ("ret", label = "coeff%i_%s_ret" % (s, ab))
    out.blank ()


def gen_complex_coeffs_inner (s, coeff, out):
    neg = coeff < 0.0
    if neg:
        coeff = -coeff
    icoeff = int (round (coeff * (1 << coeff_bits)))
    if icoeff == 2 << coeff_bits:
        icoeff -= 1

    b = 1 << coeff_bits
    cycles = 0
    if (icoeff & b) != 0:
        out.instr ("%s     vx, tx" % ("sub" if neg else "add"))
        out.instr ("%s     vy, ty" % ("sub" if neg else "add"))
        icoeff -= b
        
    c = 0
    while b != 0:
        while (icoeff & b) == 0:
            c += 1
            b >>= 1
            if b == 0:
                break
        if b == 0:
            break
        d = 0
        while (icoeff & b) != 0:
            d += 1
            b >>= 1

        if d == 1:
            out.instr ("sar     tx, #%i" % c)
            out.instr ("sar     ty, #%i" % c)
            out.instr ("%s     vx, tx" % ("sub" if neg else "add"))
            out.instr ("%s     vy, ty" % ("sub" if neg else "add"))

        else:
            if c > 1:
                out.instr ("sar     tx, #%i" % (c-1))
                out.instr ("sar     ty, #%i" % (c-1))

            out.instr ("%s     vx, tx" % ("sub" if neg else "add"))
            out.instr ("%s     vy, ty" % ("sub" if neg else "add"))
            out.instr ("sar     tx, #%i" % d)
            out.instr ("sar     ty, #%i" % d)
            out.instr ("%s     vx, tx" % ("add" if neg else "sub"))
            out.instr ("%s     vy, ty" % ("add" if neg else "sub"))
                
        c = 1

    return cycles



class Stage:
    def generate (self, out):
        raise Exception ("unimplemented")

    def process (self, sample):
        raise Exception ("unimplemented")

    def max_gain (self):
        filt = ZerosPolesGain (self.z, self.p, 1.0)
        return max_gain (filt)

stage = 1


class OnePoleStage (Stage):
    def __init__ (self, z1, p1, bits = None):
        global stage
        assert z1.imag == 0 and p1.imag == 0
        self.z = [ z1 ]
        self.p = [ p1 ]
        self.b1 = quantize ((- p1).real, bits)
        self.a1 = quantize ((- z1).real, bits)
        self.w0 = 0.0
        self.w1 = 0.0
        self.stage = stage
        stage += 1

    def process (self, sample):
        self.w1 = self.w0
        self.w0 = sample - self.b1 * self.w1 
        return self.w0 + self.a1 * self.w1


    def generate (self, out):
        print ("Onepole Stage %s max_gain %s" % (self.stage, self.max_gain()))
        simple = isclose (abs (self.a1), 1.0)
        if self.b1.imag != 0:
            raise ValueError ("Not a real pole")
        out.blank ()
        out.big_comment ("---------------------------------------------")
        out.big_comment ("Filter stage, one pole, z = %s, p = %s" % (self.z, self.p))
        out.blank ()
        gen_coeffs (self.stage, 'b', 1, self.b1.real, out)
        if not simple:
            gen_coeffs (self.stage, 'a', 1, self.a1.real, out)
        out.label ("stage%i" % self.stage)
        out.instr ("neg     t, w%i1" % self.stage)
        out.instr ("call    #coeff%i_b1" % self.stage, comment = "value -= w%i1 * %s" % (self.stage, self.b1.real))
        out.instr ("mov     w%i0, value" % self.stage, comment = "w%i0 = value" % self.stage)
        if simple:
            opcode = "sub" if self.a1.real < 0.0 else "add"
            out.instr ("%s     value, w%i1", (opcode, self.stage), comment = "value = w%i0 + w%i1" % (self.stage, self.stage))
        else:
            out.instr ("mov     t, w%i1" % self.stage)
            out.instr ("call    #coeff%i_a1" % self.stage, comment = "value += w%i1 * %s" % (self.stage, self.ba.real))
            
        out.instr ("mov     w%i1, w%i0" % (self.stage, self.stage), comment = "w%i1 <- w%i0" % (self.stage, self.stage))
        out.instr ("ret", label = "stage%i_ret" % self.stage)
        cycles += 4 * 6
        out.blank ()

        out.instr ("long    0", label = "w%i0" % self.stage)
        out.instr ("long    0", label = "w%i1" % self.stage)


class OnePoleComplexStage (Stage):
    def __init__ (self, z1, p1, bits = None):  # bits not used yet
        global stage
        self.z = [ z1 ]
        self.p = [ p1 ]
        self.b1 = - (quantize (p1.real, bits) + quantize(p1.imag, bits) * 1j)
        self.a1 = - (quantize (z1.real, bits) + quantize(z1.imag, bits) * 1j)
        self.w0 = 0.0
        self.w1 = 0.0
        self.stage = stage
        stage += 1

    def process (self, sample):
        self.w1 = self.w0
        self.w0 = sample - self.b1 * self.w1 
        return self.w0 + self.a1 * self.w1

    def generate (self, out):
        print ("Complex Stage %s max_gain %s" % (self.stage, self.max_gain()))
        out.blank ()
        out.big_comment ("---------------------------------------------")
        out.big_comment ("Filter stage, one complex pole, z = %s, p = %s" % (self.z, self.p))
        out.blank ()
        gen_complex_coeffs (self.stage, 'b', self.b1, out)
        gen_complex_coeffs (self.stage, 'a', self.a1, out)
        out.label ("stage%i" % self.stage)
        out.instr ("neg     x, w%i1x" % self.stage)
        out.instr ("neg     y, w%i1y" % self.stage)
        out.instr ("call    #coeff%i_b" % self.stage, comment = "value -= w%i1 * %s" % (self.stage, self.b1))
        out.instr ("mov     w%i0x, vx" % self.stage, comment = "w%i0 = value" % self.stage)
        out.instr ("mov     w%i0y, vy" % self.stage)
        out.instr ("mov     x, w%i1x" % self.stage)
        out.instr ("mov     y, w%i1y" % self.stage)
        out.instr ("call    #coeff%i_a" % self.stage, comment = "value += w%i1 * %s" % (self.stage, self.a1))
            
        out.instr ("mov     w%i1x, w%i0x" % (self.stage, self.stage), comment = "w%i1 <- w%i0" % (self.stage, self.stage))
        out.instr ("mov     w%i1y, w%i0y" % (2 * (self.stage,)))
        out.instr ("ret", label = "stage%i_ret" % self.stage)
        out.blank ()

        for i in range (2):
            out.instr ("long    0", label = "w%i%sx" % (self.stage, i))
            out.instr ("long    0", label = "w%i%sy" % (self.stage, i))



class TwoPoleStage (Stage):
    def __init__ (self, z1, z2, p1, p2, bits = None):
        global stage
        assert z1.imag + z2.imag == 0
        assert p1.imag + p2.imag == 0
        
        #print ("Z1 %s" % z1)
        #print ("Z2 %s" % z2)
        self.z = [z1, z2]
        self.p = [p1, p2]
        self.b1 = quantize ((- p1 - p2).real, bits)
        self.b2 = quantize ((p1 * p2).real, bits)
        self.a1 = quantize ((- z1 - z2).real, bits)
        self.a2 = quantize ((z1 * z2).real, bits)
        self.w0 = 0.0
        self.w1 = 0.0
        self.w2 = 0.0
        self.stage = stage
        stage += 1


    def process (self, sample):
        self.w2 = self.w1
        self.w1 = self.w0
        self.w0 = sample - self.b1 * self.w1 - self.b2 * self.w2
        return self.w0 + self.a1 * self.w1 + self.a2 * self.w2


    def generate (self, out):
        print ("Twopole Stage %s max_gain %s" % (self.stage, self.max_gain()))
        simple = isclose (abs (self.a1), 2.0) and isclose (self.a2, 1.0)
        out.blank ()
        out.big_comment ("---------------------------------------------")
        out.big_comment ("Filter stage, two pole, z = %s, p = %s" % (self.z, self.p))
        out.blank ()
        gen_coeffs (self.stage, 'b', 1, self.b1.real, out)
        gen_coeffs (self.stage, 'b', 2, self.b2.real, out)
        if not simple:
            gen_coeffs (self.stage, 'a', 1, self.a1.real, out)
            gen_coeffs (self.stage, 'a', 2, self.a2.real, out)
        out.label ("stage%i" % self.stage)
        out.instr ("neg     t, w%i1" % self.stage)
        out.instr ("call    #coeff%i_b1" % self.stage, comment = "value -= w%i1 * %s" % (self.stage, self.b1.real))
        out.instr ("neg     t, w%i2" % self.stage)
        out.instr ("call    #coeff%i_b2" % self.stage, comment = "value -= w%i2 * %s" % (self.stage, self.b2.real))
        out.instr ("mov     w%i0, value" % self.stage, comment = "w%i0 = value" % self.stage)
        if simple:
            opcode = "sub" if self.a1.real < 0.0 else "add"
            out.instr ("%s     value, w%i1" % (opcode, self.stage))
            out.instr ("%s     value, w%i1" % (opcode, self.stage))
            out.instr ("add     value, w%i2" % self.stage, comment = "value = w%i0 + 2 * w%i1 + w%i2" % (3 * (self.stage,)))
        else:
            out.instr ("mov     t, w%i1" % self.stage)
            out.instr ("call    #coeff%i_a1" % self.stage, comment = "value += w%i1 * %s" % (self.stage, self.a1.real))
            out.instr ("mov     t, w%i2" % self.stage)
            out.instr ("call    #coeff%i_a2" % self.stage, comment = "value += w%i2 * %s" % (self.stage, self.a2.real))
        out.instr ("mov     w%i2, w%i1" % (self.stage, self.stage))
        out.instr ("mov     w%i1, w%i0" % (self.stage, self.stage), comment = "w%i2 <- w%i1 <- w%i0" % (3 * (self.stage,)))
        out.instr ("ret", label = "stage%i_ret" % self.stage)
        out.blank ()

        for i in range (3):
            out.instr ("long    0", label = "w%i%i" % (self.stage, i))



class GainScaler:
    def __init__ (self):
        self.left_shift_pending = 0
        self.current_scaling = 1.0

    def update_gain (self, cumul_gain):
        while cumul_gain >= self.current_scaling:
            self.left_shift_pending -= 1
            self.current_scaling *= 2.0
        while cumul_gain < 0.5 * self.current_scaling:
            self.left_shift_pending += 1
            self.current_scaling *= 0.5
        print ("rescaling for %s" % cumul_gain)

    def scale (self, gain):
        return gain / self.current_scaling

    def generate_scaling (self, complex_filt, out):
        if self.left_shift_pending != 0:
            opcode = 'sar' if self.left_shift_pending < 0 else 'shl'
            shift = abs (self.left_shift_pending)
            self.left_shift_pending = 0
            if complex_filt:
                out.instr ("%s     vx, #%i" % (opcode, shift))
                out.instr ("%s     vy, #%i" % (opcode, shift))
            else:
                out.instr ("%s     value, #%i" % (opcode, shift))

        
def generate_standard_filter_header (spin_file, out):
    out.big_comment (spin_file)
    out.print ("""
CON

OBJ


PUB Start (param_address)
  RESULT := cognew (@entry, param_address)

DAT
                ORG     0
entry
                mov     parm, PAR
                rdlong  period, parm
		add	parm, #4
		rdlong  time, parm
		add	parm, #4
	        mov     in_addr, parm
                add     parm, #4
		mov	out_addr, parm

:filtloop       WAITCNT time, period   ' synchronize
		rdlong  value, in_addr
                call    #filter
                wrlong  value, out_addr
                jmp     #:filtloop

""")

def generate_complex_filter_header (spin_file, out):
    out.big_comment (spin_file)
    out.print ("""
CON

OBJ


PUB Start (param_address)
  RESULT := cognew (@entry, param_address)

DAT
                ORG     0
entry
                mov     parm, PAR
                rdlong  period, parm
		add	parm, #4
		rdlong  time, parm
		add	parm, #4
	        mov     in_re_addr, parm
                add     parm, #4
	        mov     in_im_addr, parm
                add     parm, #4
		mov	out_re_addr, parm
                add     parm, #4
		mov	out_im_addr, parm

:filtloop       WAITCNT time, period   ' synchronize
		rdlong  vx, in_re_addr
                rdlong  vy, in_im_addr
                call    #filter
                wrlong  vx, out_re_addr
                wrlong  vy, out_im_addr
                jmp     #:filtloop

""")

def generate_standard_filter_footer (out):
    for symbol in [ "parm", "time", "period", "in_addr", "out_addr" ]:
        out.instr ("res     1", label = symbol)
    out.instr ("FIT     $1F0")

def generate_complex_filter_footer (out):
    for symbol in [ "parm", "time", "period", "in_re_addr", "in_im_addr", "out_re_addr", "out_im_addr" ]:
        out.instr ("res     1", label = symbol)
    out.instr ("FIT     $1F0")
    

def compile_propeller_filter (resp, spinfile, bits):
    raw_zeroes = resp.zeros
    raw_poles = resp.poles
    raw_gain = resp.gain
    print (max_gain (resp))
    print (max_gain (ZerosPolesGain (raw_zeroes, raw_poles, 1.0)))

    conj_zeroes, sing_zeroes, compl_zeroes = factorize_poly (raw_zeroes)
    conj_poles,  sing_poles,  compl_poles =  factorize_poly (raw_poles)

    complex_filt = False
    if len(compl_zeroes) + len(compl_poles) > 0:
        complex_filt = True
        assert len (conj_zeroes) + len (sing_zeroes) == 0
        assert len (conj_poles) + len (sing_poles) == 0

    conj_zeroes = sorted (conj_zeroes, key = lambda x: np.real(x))
    conj_poles  = sorted (conj_poles,  key = lambda x: -abs(x))

    print ("conj_zeroes %s" % conj_zeroes)
    print ("conj_poles %s" % conj_poles) # [ abs(p) for p in conj_poles])

    complete_order_2_count = min (len (conj_zeroes), len (conj_poles))
    complete_order_1_count = min (len (sing_zeroes), len (sing_poles))
    zero_2_pole_1_count = min (len (conj_zeroes) - complete_order_2_count, len (sing_poles) - complete_order_1_count)
    zero_1_pole_2_count = min (len (conj_poles) - complete_order_2_count, len (sing_zeroes) - complete_order_1_count)
    zero_2_count = len (conj_zeroes) - complete_order_2_count - zero_2_pole_1_count
    pole_2_count = len (conj_poles) - complete_order_2_count - zero_1_pole_2_count
    zero_1_count = len (sing_zeroes) - complete_order_1_count - zero_1_pole_2_count
    pole_1_count = len (sing_poles) - complete_order_1_count - zero_2_pole_1_count

    print ("order2 = %s, order1 = %s" % (complete_order_2_count, complete_order_1_count))
    print ("zero_2_pole_1_count = %s" % zero_2_pole_1_count)
    print ("zero_1_pole_2_count = %s" % zero_1_pole_2_count)
    print ("zero_2_count = %s, zero_1_count = %s" % (zero_2_count, zero_1_count))
    print ("pole_2_count = %s, pole_1_count = %s" % (pole_2_count, pole_1_count))
    
    with open (spinfile, 'w') as f:
        out = Outputter (f)
        if complex_filt:
            generate_complex_filter_header (spinfile, out)
        else:
            generate_standard_filter_header (spinfile, out)
        stages = []
        #if zero_1_count == 0 and zero_2_count == 0 and pole_1_count == 0 and pole_2_count == 0:
        if  zero_2_count == 0 and pole_1_count == 0 and pole_2_count == 0:
            if zero_2_pole_1_count == 0 and zero_1_pole_2_count == 0:
                for i in range (complete_order_2_count):
                    stages.append (TwoPoleStage (conj_zeroes[i], np.conj (conj_zeroes[i]),
                                                 conj_poles[i], np.conj (conj_poles[i]),
                                                 bits = bits))
                for i in range (complete_order_1_count):
                    stages.append (OnePoleStage (sing_zeroes[i],
                                                 sing_poles[i],
                                                 bits = bits))
            else:
                for i in range (zero_1_pole_2_count):
                    stages.append (TwoPoleStage (sing_zeroes[i], 0.0,
                                                 conj_poles[i], np.conj (conj_poles[i]),
                                                 bits = bits))

        for i in range (max (len (compl_poles), len (compl_zeroes))):
            stages.append (OnePoleComplexStage (compl_zeroes[i],
                                                compl_poles[i],
                                                bits = bits))

        print ("Have %s stages" % len(stages))
        # Generate driver with scaling for gain compensation
        out.label ("filter")

        cumul_zeroes, cumul_poles = [], []
        scaler = GainScaler () #left_shift, gain_scaling = 0, 1.0
        for stage in stages:
            cumul_poles  += stage.p
            cumulative_gain = max_gain (ZerosPolesGain (cumul_zeroes, cumul_poles, 1.0))

            scaler.update_gain (cumulative_gain)
            scaler.generate_scaling (complex_filt, out)

            out.instr ("call    #stage%i" % stage.stage,
                       comment = "stage gain = %s, cumul actual gain = %s" % (stage.max_gain(),
                                                                              scaler.scale (cumulative_gain)))
            out.comment ("poles only = %s" % max_gain (ZerosPolesGain ([], stage.p, 1.0)))

            cumul_zeroes += stage.z
            cumulative_gain = max_gain (ZerosPolesGain (cumul_zeroes, cumul_poles, 1.0))
            scaler.update_gain (cumulative_gain)

        scaler.generate_scaling (complex_filt, out)

        out.instr ("ret", label = "filter_ret", comment = "cumul actual gain = %s\n" % scaler.scale (cumulative_gain))
        if complex_filt:
            out.instr ("long    0", label = "vx")
            out.instr ("long    0", label = "vy")
        else:
            out.instr ("long    0", label = "value")

        # Generate stages
        for stage in stages:
            stage.generate (out)

        out.blank ()
        if complex_filt:
            for symbol in [ "x", "y", "tx", "ty" ]:
                out.instr ("res     1", label = symbol)
        else:
            out.instr ("res     1", label = "t")

                
        out.big_comment  ("%s bits per coefficient" % bits)
        out.big_comment (cycles_desc (out.cycles()))
        if complex_filt:
            generate_complex_filter_footer (out)
        else:
            generate_standard_filter_footer (out)

"""
////////////////////////////////////////////////////////////////////////////////////////////
//                                TERMS OF USE: MIT License
////////////////////////////////////////////////////////////////////////////////////////////
// Permission is hereby granted, free of charge, to any person obtaining a copy of this 
// software and associated documentation files (the "Software"), to deal in the Software 
// without restriction, including without limitation the rights to use, copy, modify, merge,
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit 
// persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or
// substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
// DEALINGS IN THE SOFTWARE.
////////////////////////////////////////////////////////////////////////////////////////////
"""
